__version__ = "0.1.7rc1"
from qibo.config import set_batch_size, get_batch_size
from qibo.config import set_metropolis_threshold, get_metropolis_threshold
from qibo.backends import set_backend, get_backend
from qibo.backends import set_precision, get_precision
from qibo.backends import set_device, get_device
from qibo.backends import set_threads, get_threads
from qibo.backends import numpy_matrices as matrices
from qibo.backends import K
from qibo import callbacks, gates, hamiltonians, models
from qibo import parallel, optimizers, solvers
